/*
 * bsc1184952_fuse_file
 *
 * Fix for CVE-2020-36322, bsc#1184952 (fs/fuse/file.c part)
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "bsc1184952_common.h"

/* klp-ccp: from fs/fuse/fuse_i.h */
static void (*klpe_fuse_read_args_fill)(struct fuse_io_args *ia, struct file *file, loff_t pos,
			 size_t count, int opcode);

int klpp_fuse_open_common(struct inode *inode, struct file *file, bool isdir);

static void (*klpe_fuse_finish_open)(struct inode *inode, struct file *file);

static int (*klpe_fuse_fsync_common)(struct file *file, loff_t start, loff_t end,
		      int datasync, int opcode);

static ssize_t (*klpe_fuse_simple_request)(struct fuse_conn *fc, struct fuse_args *args);
static int (*klpe_fuse_simple_background)(struct fuse_conn *fc, struct fuse_args *args,
			   gfp_t gfp_flags);

static void (*klpe_fuse_invalidate_attr)(struct inode *inode);

static void (*klpe_fuse_invalidate_atime)(struct inode *inode);

static int (*klpe_fuse_allow_current_process)(struct fuse_conn *fc);

static u64 (*klpe_fuse_lock_owner_id)(struct fuse_conn *fc, fl_owner_t id);

static int (*klpe_fuse_update_attributes)(struct inode *inode, struct file *file);

static void (*klpe_fuse_set_nowrite)(struct inode *inode);
static void (*klpe_fuse_release_nowrite)(struct inode *inode);

static int (*klpe_fuse_do_open)(struct fuse_conn *fc, u64 nodeid, struct file *file,
		 bool isdir);

static ssize_t (*klpe_fuse_direct_io)(struct fuse_io_priv *io, struct iov_iter *iter,
		       loff_t *ppos, int flags);
static long (*klpe_fuse_do_ioctl)(struct file *file, unsigned int cmd, unsigned long arg,
		   unsigned int flags);
long klpp_fuse_ioctl_common(struct file *file, unsigned int cmd,
		       unsigned long arg, unsigned int flags);

static bool (*klpe_fuse_write_update_size)(struct inode *inode, loff_t pos);

/* klp-ccp: from fs/fuse/file.c */
#include <linux/pagemap.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/compat.h>
#include <linux/uio.h>
#include <linux/fs.h>

struct fuse_release_args {
	struct fuse_args args;
	struct fuse_release_in inarg;
	struct inode *inode;
};

static struct fuse_file *fuse_file_get(struct fuse_file *ff)
{
	refcount_inc(&ff->count);
	return ff;
}

static void (*klpe_fuse_release_end)(struct fuse_conn *fc, struct fuse_args *args,
			     int error);

static void klpr_fuse_file_put(struct fuse_file *ff, bool sync, bool isdir)
{
	if (refcount_dec_and_test(&ff->count)) {
		struct fuse_args *args = &ff->release_args->args;

		if (isdir ? ff->fc->no_opendir : ff->fc->no_open) {
			/* Do nothing when client does not implement 'open' */
			(*klpe_fuse_release_end)(ff->fc, args, 0);
		} else if (sync) {
			(*klpe_fuse_simple_request)(ff->fc, args);
			(*klpe_fuse_release_end)(ff->fc, args, 0);
		} else {
			args->end = (*klpe_fuse_release_end);
			if ((*klpe_fuse_simple_background)(ff->fc, args,
						   GFP_KERNEL | __GFP_NOFAIL))
				(*klpe_fuse_release_end)(ff->fc, args, -ENOTCONN);
		}
		kfree(ff);
	}
}

int klpp_fuse_open_common(struct inode *inode, struct file *file, bool isdir)
{
	struct fuse_conn *fc = get_fuse_conn(inode);
	int err;
	bool is_wb_truncate = (file->f_flags & O_TRUNC) &&
			  fc->atomic_o_trunc &&
			  fc->writeback_cache;

	/*
	 * Fix CVE-2020-36322
	 *  +3 lines
	 */
	if (klpp_fuse_is_bad(inode))
		return -EIO;

	err = generic_file_open(inode, file);
	if (err)
		return err;

	if (is_wb_truncate) {
		inode_lock(inode);
		(*klpe_fuse_set_nowrite)(inode);
	}

	err = (*klpe_fuse_do_open)(fc, get_node_id(inode), file, isdir);

	if (!err)
		(*klpe_fuse_finish_open)(inode, file);

	if (is_wb_truncate) {
		(*klpe_fuse_release_nowrite)(inode);
		inode_unlock(inode);
	}

	return err;
}

struct fuse_writepage_args {
	struct fuse_io_args ia;
	struct list_head writepages_entry;
	struct list_head queue_entry;
	struct fuse_writepage_args *next;
	struct inode *inode;
};

static void (*klpe_fuse_wait_on_page_writeback)(struct inode *inode, pgoff_t index);

static void klpr_fuse_sync_writes(struct inode *inode)
{
	(*klpe_fuse_set_nowrite)(inode);
	(*klpe_fuse_release_nowrite)(inode);
}

int klpp_fuse_flush(struct file *file, fl_owner_t id)
{
	struct inode *inode = file_inode(file);
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_file *ff = file->private_data;
	struct fuse_flush_in inarg;
	FUSE_ARGS(args);
	int err;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		return -EIO;

	err = 0;
	if (fc->no_flush)
		goto inval_attr_out;

	err = write_inode_now(inode, 1);
	if (err)
		return err;

	inode_lock(inode);
	klpr_fuse_sync_writes(inode);
	inode_unlock(inode);

	err = filemap_check_errors(file->f_mapping);
	if (err)
		return err;

	memset(&inarg, 0, sizeof(inarg));
	inarg.fh = ff->fh;
	inarg.lock_owner = (*klpe_fuse_lock_owner_id)(fc, id);
	args.opcode = FUSE_FLUSH;
	args.nodeid = get_node_id(inode);
	args.in_numargs = 1;
	args.in_args[0].size = sizeof(inarg);
	args.in_args[0].value = &inarg;
	args.force = true;

	err = (*klpe_fuse_simple_request)(fc, &args);
	if (err == -ENOSYS) {
		fc->no_flush = 1;
		err = 0;
	}

inval_attr_out:
	/*
	 * In memory i_blocks is not maintained by fuse, if writeback cache is
	 * enabled, i_blocks from cached attr may not be accurate.
	 */
	if (!err && fc->writeback_cache)
		(*klpe_fuse_invalidate_attr)(inode);
	return err;
}

int klpp_fuse_fsync(struct file *file, loff_t start, loff_t end,
		      int datasync)
{
	struct inode *inode = file->f_mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	int err;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		return -EIO;

	inode_lock(inode);

	/*
	 * Start writeback against all dirty pages of the inode, then
	 * wait for all outstanding writes, before sending the FSYNC
	 * request.
	 */
	err = file_write_and_wait_range(file, start, end);
	if (err)
		goto out;

	klpr_fuse_sync_writes(inode);

	/*
	 * Due to implementation of fuse writeback
	 * file_write_and_wait_range() does not catch errors.
	 * We have to do this directly after fuse_sync_writes()
	 */
	err = file_check_and_advance_wb_err(file);
	if (err)
		goto out;

	err = sync_inode_metadata(inode, 1);
	if (err)
		goto out;

	if (fc->no_fsync)
		goto out;

	err = (*klpe_fuse_fsync_common)(file, start, end, datasync, FUSE_FSYNC);
	if (err == -ENOSYS) {
		fc->no_fsync = 1;
		err = 0;
	}
out:
	inode_unlock(inode);

	return err;
}

static struct fuse_io_args *(*klpe_fuse_io_alloc)(struct fuse_io_priv *io,
					  unsigned int npages);

static void fuse_io_free(struct fuse_io_args *ia)
{
	kfree(ia->ap.pages);
	kfree(ia);
}

static void fuse_read_update_size(struct inode *inode, loff_t size,
				  u64 attr_ver)
{
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_inode *fi = get_fuse_inode(inode);

	spin_lock(&fi->lock);
	if (attr_ver == fi->attr_version && size < inode->i_size &&
	    !test_bit(FUSE_I_SIZE_UNSTABLE, &fi->state)) {
		fi->attr_version = atomic64_inc_return(&fc->attr_version);
		i_size_write(inode, size);
	}
	spin_unlock(&fi->lock);
}

static void fuse_short_read(struct inode *inode, u64 attr_ver, size_t num_read,
			    struct fuse_args_pages *ap)
{
	struct fuse_conn *fc = get_fuse_conn(inode);

	if (fc->writeback_cache) {
		/*
		 * A hole in a file. Some data after the hole are in page cache,
		 * but have not reached the client fs yet. So, the hole is not
		 * present there.
		 */
		int i;
		int start_idx = num_read >> PAGE_SHIFT;
		size_t off = num_read & (PAGE_SIZE - 1);

		for (i = start_idx; i < ap->num_pages; i++) {
			zero_user_segment(ap->pages[i], off, PAGE_SIZE);
			off = 0;
		}
	} else {
		loff_t pos = page_offset(ap->pages[0]) + num_read;
		fuse_read_update_size(inode, pos, attr_ver);
	}
}

static int klpr_fuse_do_readpage(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	loff_t pos = page_offset(page);
	struct fuse_page_desc desc = { .length = PAGE_SIZE };
	struct fuse_io_args ia = {
		.ap.args.page_zeroing = true,
		.ap.args.out_pages = true,
		.ap.num_pages = 1,
		.ap.pages = &page,
		.ap.descs = &desc,
	};
	ssize_t res;
	u64 attr_ver;

	/*
	 * Page writeback can extend beyond the lifetime of the
	 * page-cache page, so make sure we read a properly synced
	 * page.
	 */
	(*klpe_fuse_wait_on_page_writeback)(inode, page->index);

	attr_ver = fuse_get_attr_version(fc);

	/* Don't overflow end offset */
	if (pos + (desc.length - 1) == LLONG_MAX)
		desc.length--;

	(*klpe_fuse_read_args_fill)(&ia, file, pos, desc.length, FUSE_READ);
	res = (*klpe_fuse_simple_request)(fc, &ia.ap.args);
	if (res < 0)
		return res;
	/*
	 * Short read means EOF.  If file size is larger, truncate it
	 */
	if (res < desc.length)
		fuse_short_read(inode, attr_ver, res, &ia.ap);

	SetPageUptodate(page);

	return 0;
}

int klpp_fuse_readpage(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	int err;

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		goto out;

	err = klpr_fuse_do_readpage(file, page);
	(*klpe_fuse_invalidate_atime)(inode);
 out:
	unlock_page(page);
	return err;
}

static void (*klpe_fuse_readpages_end)(struct fuse_conn *fc, struct fuse_args *args,
			       int err);

static void klpr_fuse_send_readpages(struct fuse_io_args *ia, struct file *file)
{
	struct fuse_file *ff = file->private_data;
	struct fuse_conn *fc = ff->fc;
	struct fuse_args_pages *ap = &ia->ap;
	loff_t pos = page_offset(ap->pages[0]);
	size_t count = ap->num_pages << PAGE_SHIFT;
	ssize_t res;
	int err;

	ap->args.out_pages = true;
	ap->args.page_zeroing = true;
	ap->args.page_replace = true;

	/* Don't overflow end offset */
	if (pos + (count - 1) == LLONG_MAX) {
		count--;
		ap->descs[ap->num_pages - 1].length--;
	}
	WARN_ON((loff_t) (pos + count) < 0);

	(*klpe_fuse_read_args_fill)(ia, file, pos, count, FUSE_READ);
	ia->read.attr_ver = fuse_get_attr_version(fc);
	if (fc->async_read) {
		ia->ff = fuse_file_get(ff);
		ap->args.end = (*klpe_fuse_readpages_end);
		err = (*klpe_fuse_simple_background)(fc, &ap->args, GFP_KERNEL);
		if (!err)
			return;
	} else {
		res = (*klpe_fuse_simple_request)(fc, &ap->args);
		err = res < 0 ? res : 0;
	}
	(*klpe_fuse_readpages_end)(fc, &ap->args, err);
}

struct fuse_fill_data {
	struct fuse_io_args *ia;
	struct file *file;
	struct inode *inode;
	unsigned int nr_pages;
	unsigned int max_pages;
};

static int (*klpe_fuse_readpages_fill)(void *_data, struct page *page);

int klpp_fuse_readpages(struct file *file, struct address_space *mapping,
			  struct list_head *pages, unsigned nr_pages)
{
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_fill_data data;
	int err;

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		goto out;

	data.file = file;
	data.inode = inode;
	data.nr_pages = nr_pages;
	data.max_pages = min_t(unsigned int, nr_pages, fc->max_pages);
;
	data.ia = (*klpe_fuse_io_alloc)(NULL, data.max_pages);
	err = -ENOMEM;
	if (!data.ia)
		goto out;

	err = read_cache_pages(mapping, pages, (*klpe_fuse_readpages_fill), &data);
	if (!err) {
		if (data.ia->ap.num_pages)
			klpr_fuse_send_readpages(data.ia, file);
		else
			fuse_io_free(data.ia);
	}
out:
	return err;
}

static ssize_t klpr_fuse_cache_read_iter(struct kiocb *iocb, struct iov_iter *to)
{
	struct inode *inode = iocb->ki_filp->f_mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);

	/*
	 * In auto invalidate mode, always update attributes on read.
	 * Otherwise, only update if we attempt to read past EOF (to ensure
	 * i_size is up to date).
	 */
	if (fc->auto_inval_data ||
	    (iocb->ki_pos + iov_iter_count(to) > i_size_read(inode))) {
		int err;
		err = (*klpe_fuse_update_attributes)(inode, iocb->ki_filp);
		if (err)
			return err;
	}

	return generic_file_read_iter(iocb, to);
}

static ssize_t (*klpe_fuse_perform_write)(struct kiocb *iocb,
				  struct address_space *mapping,
				  struct iov_iter *ii, loff_t pos);

static ssize_t klpr_fuse_cache_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	struct file *file = iocb->ki_filp;
	struct address_space *mapping = file->f_mapping;
	ssize_t written = 0;
	ssize_t written_buffered = 0;
	struct inode *inode = mapping->host;
	ssize_t err;
	loff_t endbyte = 0;

	if (get_fuse_conn(inode)->writeback_cache) {
		/* Update size (EOF optimization) and mode (SUID clearing) */
		err = (*klpe_fuse_update_attributes)(mapping->host, file);
		if (err)
			return err;

		return generic_file_write_iter(iocb, from);
	}

	inode_lock(inode);

	/* We can write back this queue in page reclaim */
	current->backing_dev_info = inode_to_bdi(inode);

	err = generic_write_checks(iocb, from);
	if (err <= 0)
		goto out;

	err = file_remove_privs(file);
	if (err)
		goto out;

	err = file_update_time(file);
	if (err)
		goto out;

	if (iocb->ki_flags & IOCB_DIRECT) {
		loff_t pos = iocb->ki_pos;
		written = generic_file_direct_write(iocb, from);
		if (written < 0 || !iov_iter_count(from))
			goto out;

		pos += written;

		written_buffered = (*klpe_fuse_perform_write)(iocb, mapping, from, pos);
		if (written_buffered < 0) {
			err = written_buffered;
			goto out;
		}
		endbyte = pos + written_buffered - 1;

		err = filemap_write_and_wait_range(file->f_mapping, pos,
						   endbyte);
		if (err)
			goto out;

		invalidate_mapping_pages(file->f_mapping,
					 pos >> PAGE_SHIFT,
					 endbyte >> PAGE_SHIFT);

		written += written_buffered;
		iocb->ki_pos = pos + written_buffered;
	} else {
		written = (*klpe_fuse_perform_write)(iocb, mapping, from, iocb->ki_pos);
		if (written >= 0)
			iocb->ki_pos += written;
	}
out:
	current->backing_dev_info = NULL;
	inode_unlock(inode);
	if (written > 0)
		written = generic_write_sync(iocb, written);

	return written ? written : err;
}

static ssize_t klpr___fuse_direct_read(struct fuse_io_priv *io,
				  struct iov_iter *iter,
				  loff_t *ppos)
{
	ssize_t res;
	struct inode *inode = file_inode(io->iocb->ki_filp);

	res = (*klpe_fuse_direct_io)(io, iter, ppos, 0);

	(*klpe_fuse_invalidate_atime)(inode);

	return res;
}

static ssize_t (*klpe_fuse_direct_IO)(struct kiocb *iocb, struct iov_iter *iter);

static ssize_t klpr_fuse_direct_read_iter(struct kiocb *iocb, struct iov_iter *to)
{
	ssize_t res;

	if (!is_sync_kiocb(iocb) && iocb->ki_flags & IOCB_DIRECT) {
		res = (*klpe_fuse_direct_IO)(iocb, to);
	} else {
		struct fuse_io_priv io = FUSE_IO_PRIV_SYNC(iocb);

		res = klpr___fuse_direct_read(&io, to, &iocb->ki_pos);
	}

	return res;
}

static ssize_t klpr_fuse_direct_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	struct inode *inode = file_inode(iocb->ki_filp);
	struct fuse_io_priv io = FUSE_IO_PRIV_SYNC(iocb);
	ssize_t res;

	/* Don't allow parallel writes to the same file */
	inode_lock(inode);
	res = generic_write_checks(iocb, from);
	if (res > 0) {
		if (!is_sync_kiocb(iocb) && iocb->ki_flags & IOCB_DIRECT) {
			res = (*klpe_fuse_direct_IO)(iocb, from);
		} else {
			res = (*klpe_fuse_direct_io)(&io, from, &iocb->ki_pos,
					     FUSE_DIO_WRITE);
		}
	}
	(*klpe_fuse_invalidate_attr)(inode);
	if (res > 0)
		(*klpe_fuse_write_update_size)(inode, iocb->ki_pos);
	inode_unlock(inode);

	return res;
}

ssize_t klpp_fuse_file_read_iter(struct kiocb *iocb, struct iov_iter *to)
{
	struct file *file = iocb->ki_filp;
	struct fuse_file *ff = file->private_data;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(file_inode(file)))
		return -EIO;

	if (!(ff->open_flags & FOPEN_DIRECT_IO))
		return klpr_fuse_cache_read_iter(iocb, to);
	else
		return klpr_fuse_direct_read_iter(iocb, to);
}

ssize_t klpp_fuse_file_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	struct file *file = iocb->ki_filp;
	struct fuse_file *ff = file->private_data;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(file_inode(file)))
		return -EIO;

	if (!(ff->open_flags & FOPEN_DIRECT_IO))
		return klpr_fuse_cache_write_iter(iocb, from);
	else
		return klpr_fuse_direct_write_iter(iocb, from);
}

struct fuse_fill_wb_data {
	struct fuse_writepage_args *wpa;
	struct fuse_file *ff;
	struct inode *inode;
	struct page **orig_pages;
	unsigned int max_pages;
};

static void (*klpe_fuse_writepages_send)(struct fuse_fill_wb_data *data);

static int (*klpe_fuse_writepages_fill)(struct page *page,
		struct writeback_control *wbc, void *_data);

int klpp_fuse_writepages(struct address_space *mapping,
			   struct writeback_control *wbc)
{
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_fill_wb_data data;
	int err;

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		goto out;

	data.inode = inode;
	data.wpa = NULL;
	data.ff = NULL;

	err = -ENOMEM;
	data.orig_pages = kcalloc(fc->max_pages,
				  sizeof(struct page *),
				  GFP_NOFS);
	if (!data.orig_pages)
		goto out;

	err = write_cache_pages(mapping, wbc, (*klpe_fuse_writepages_fill), &data);
	if (data.wpa) {
		WARN_ON(!data.wpa->ia.ap.num_pages);
		(*klpe_fuse_writepages_send)(&data);
	}
	if (data.ff)
		klpr_fuse_file_put(data.ff, false, false);

	kfree(data.orig_pages);
out:
	return err;
}

long klpp_fuse_ioctl_common(struct file *file, unsigned int cmd,
		       unsigned long arg, unsigned int flags)
{
	struct inode *inode = file_inode(file);
	struct fuse_conn *fc = get_fuse_conn(inode);

	if (!(*klpe_fuse_allow_current_process)(fc))
		return -EACCES;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (klpp_fuse_is_bad(inode))
		return -EIO;

	return (*klpe_fuse_do_ioctl)(file, cmd, arg, flags);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1184952.h"
#include "../kallsyms_relocs.h"

#define LIVEPATCHED_MODULE "fuse"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "fuse_simple_background", (void *)&klpe_fuse_simple_background,
	  "fuse" },
	{ "fuse_simple_request", (void *)&klpe_fuse_simple_request, "fuse" },
	{ "fuse_read_args_fill", (void *)&klpe_fuse_read_args_fill, "fuse" },
	{ "fuse_finish_open", (void *)&klpe_fuse_finish_open, "fuse" },
	{ "fuse_fsync_common", (void *)&klpe_fuse_fsync_common, "fuse" },
	{ "fuse_invalidate_attr", (void *)&klpe_fuse_invalidate_attr, "fuse" },
	{ "fuse_invalidate_atime", (void *)&klpe_fuse_invalidate_atime,
	  "fuse" },
	{ "fuse_allow_current_process",
	  (void *)&klpe_fuse_allow_current_process, "fuse" },
	{ "fuse_lock_owner_id", (void *)&klpe_fuse_lock_owner_id, "fuse" },
	{ "fuse_update_attributes", (void *)&klpe_fuse_update_attributes,
	  "fuse" },
	{ "fuse_set_nowrite", (void *)&klpe_fuse_set_nowrite, "fuse" },
	{ "fuse_release_nowrite", (void *)&klpe_fuse_release_nowrite, "fuse" },
	{ "fuse_do_open", (void *)&klpe_fuse_do_open, "fuse" },
	{ "fuse_direct_io", (void *)&klpe_fuse_direct_io, "fuse" },
	{ "fuse_do_ioctl", (void *)&klpe_fuse_do_ioctl, "fuse" },
	{ "fuse_write_update_size", (void *)&klpe_fuse_write_update_size,
	  "fuse" },
	{ "fuse_readpages_end", (void *)&klpe_fuse_readpages_end, "fuse" },
	{ "fuse_writepages_fill", (void *)&klpe_fuse_writepages_fill, "fuse" },
	{ "fuse_writepages_send", (void *)&klpe_fuse_writepages_send, "fuse" },
	{ "fuse_io_alloc", (void *)&klpe_fuse_io_alloc, "fuse" },
	{ "fuse_readpages_fill", (void *)&klpe_fuse_readpages_fill, "fuse" },
	{ "fuse_perform_write", (void *)&klpe_fuse_perform_write, "fuse" },
	{ "fuse_direct_IO", (void *)&klpe_fuse_direct_IO, "fuse" },
	{ "fuse_wait_on_page_writeback",
	  (void *)&klpe_fuse_wait_on_page_writeback, "fuse" },
	{ "fuse_release_end", (void *)&klpe_fuse_release_end, "fuse" },
};

static int livepatch_bsc1184952_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LIVEPATCHED_MODULE))
		return 0;

	mutex_lock(&module_mutex);
	ret = __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	mutex_unlock(&module_mutex);
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1184952_module_nb = {
	.notifier_call = livepatch_bsc1184952_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1184952_fuse_file_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(LIVEPATCHED_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&livepatch_bsc1184952_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1184952_fuse_file_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1184952_module_nb);
}
